/*
 * EXAMPLE 1
 * -------------------------------------------
 * MIGUEL TORRES TORRITI
 * -------------------------------------------
 * This example shows how to:
 *  1. Load an image using the new-style image classes declared in cxcore.hpp, in particular (CvImage), a wrapper for IplImage.
 *     It is to be noted that:
 *     a. To access the elements of the CvImage object either type cast the CvImage to IplImage or
 *        use C++ '->' operator, i.e. the following are equivalent:
 * 		 W=((IplImage*) img[0])->width;
 * 		 W=img[0].width();
 * 		 W=img->width();
 *     b. There is also a very similar matrix class (CvMatrix) - a wrapper for CvMat.
 *  2. Perform basic matrix operations: Create/Save/Load.
 *  3. Perform advanced matrix operations: Allocating/initializing a matrix in an external function.  
 *  4. Manipulate basic structures and data types, such as CvMat, CvScalar.
 *  5. Create/copy an image.
 *  6. Use a trackbar.
 *  7. Display an image.
 *  8. Draw a primitive.
 *  9. Apply an image processing function from the OpenCV library.
 * 10. Read and modify image pixels.
 * 11. Save an image.
 * To run type:
 *   ocvs <image filename>
 * By default ocvs loads lena.jpg, if not found it will display and error message.
 *
 */

#include "cv.h"      // use " " instead of < > to get list of class/struct members
#include "highgui.h" // use " " instead of < > to get list of class/struct members
#include <stdio.h>
#include <math.h>

// Define global variables
#define pi 3.1415926535897932384626433832795

CvImage gray, edge, img2;
CvRNG rng = cvRNG(-1);
char wndname[] = "Edge";
char tbarname[] = "Threshold";
int edge_thresh = 1;

// Define a trackbar callback
void on_trackbar(int h)
{	int hh, ww;
	
	//cvSmooth( gray, edge, CV_BLUR, 3, 3, 0, 0 );
	//cvNot( gray, edge );

    // Run the edge detector on grayscale
    cvCanny(gray, edge, (float)edge_thresh, (float)edge_thresh*3.0, 3);

	hh=((IplImage*) edge)->height;
	ww=((IplImage*) edge)->width;
	cvLine( edge, cvPoint(0,0), cvPoint(ww,hh), CV_RGB(255,255,255), 1, 8, 0 );

	edge.show(wndname);
	
	CvScalar Iv;

	// Reading and printing image intensity levels
	printf("Showing grayscale image intensity levels (format: Pixel Index - Intensity)\n");
	for(int idx0 = 0; idx0 < 5; idx0++ )
	{
		Iv = cvGet1D( gray, idx0 );
		printf("%d - %d\n", idx0, (int) Iv.val[0]);
	}

	printf("Showing color image intensity levels (format: Pixel Index - R : G : B)\n");
	for(int idx0 = 0; idx0 < 5; idx0++ )
	{
		Iv = cvGet1D( img2, idx0 );
		printf("%d - %d : %d : %d\n", idx0, (int) Iv.val[2], (int) Iv.val[1], (int) Iv.val[0]);
	}

	// Modifying image intensinty levels 
	//Iv.val[0]=0;   //b
	//Iv.val[1]=0;   //g
	//Iv.val[2]=255; //r
	Iv = CV_RGB(255,0,0);

	for(int idx0 = ww*(hh/2-20)+ww/2-20; idx0 < ww*(hh/2-20)+ww/2+20; idx0++ )
	{
		cvSet1D(img2, idx0, Iv); //idx0+1, idx0+2, Iv );	
	}
	for(int idx0 = ww*(hh/2+20)+ww/2-20; idx0 < ww*(hh/2+20)+ww/2+20; idx0++ )
	{
		cvSet1D(img2, idx0, Iv); //idx0+1, idx0+2, Iv );	
	}
	for(int idx0 = ww*(hh/2-20)+ww/2-20; idx0 < ww*(hh/2+20)+ww/2-20; idx0+=ww )
	{
		cvSet1D(img2, idx0, Iv); //idx0+1, idx0+2, Iv );	
	}
	for(int idx0 = ww*(hh/2-20)+ww/2+20; idx0 < ww*(hh/2+20)+ww/2+20; idx0+=ww )
	{
		cvSet1D(img2, idx0, Iv); //idx0+1, idx0+2, Iv );	
	}
	img2.show( "Image Copy" );
}

struct complex {
	double real;
	double imaginary;
};

struct data {
	int n;
	complex* z;
	double* vector;
	CvScalar q;
};

void mfun(void *mdata, CvScalar x0, CvScalar *x)
{
	int Nx = ((struct data *) mdata)->n;

	for(int i=0; i<Nx; i++)
	{
		(*x).val[i]=x0.val[i] + ((struct data *) mdata)->q.val[i] + ((struct data *) mdata)->vector[i];
		printf("x0(%d)=%.2f, x(%d)=%.2f\n",i,x0.val[i],i,(*x).val[i]);
	}
	printf("Initial z = %.2f+i%.2f\n", ((struct data *) mdata)->z->real, (*(((struct data *) mdata)->z)).imaginary);
	((struct data *) mdata)->z->real=2.0;
	((struct data *) mdata)->z[0].imaginary=pi;
	(*((struct data *) mdata)->z).real = (((struct data *) mdata)->z)->real + 3.0;
	(*(((struct data *) mdata)->z)).imaginary = ((struct data *) mdata)->z->imaginary + pi;
	printf("Final z = %.2f+i%.2f\n", ((struct data *) mdata)->z->real, ((struct data *) mdata)->z->imaginary);
}

void mproc(void)
{
	int dim = 3;
	double vec[] = { 1.0, pi, 5};
	complex z1 = {1.0, sqrt(2.0)};
	struct data d = { dim, &z1, vec, cvScalar(3.0, 8.0, -1.0) };
	CvScalar x0 = cvScalar(12.0, -21.0, 5.0), x; 
	mfun(&d, x0, &x);
}


void matrixcreatesaveload(void)
{
	// Create a matrix
	int Nx = 5, Ny = 3;
	CvMat*	m_xy = cvCreateMat(Ny, Nx, CV_32F);
	//cvCreateData(m_xy); // Do not use! u_xy already allocated by cvCreateMat
	cvSetZero(m_xy);

	printf("Matrix m_xy rows: %d, cols: %d\n", m_xy->rows, m_xy->cols);
	//cvmSet(m_xy, m_xy->rows-1, m_xy->cols-1, pi);

	for(int i=0; i<Ny; i++)
	{
		for(int j=0; j<Nx; j++)
		{
			cvmSet(m_xy, i, j, pi*sin(2.0*pi*(double) (i+j)/(double) (Ny+Nx-2)) );
			printf("(%d,%d,%.2f) ", i, j, cvmGet(m_xy, i, j) );
		}
		printf("\n");
     }

	// Save a matrix using cvSave
	cvSave( "m_xy.xml", m_xy, "m_xy", NULL, cvAttrList(0,0) );// cannot use non-alpha characters in object descriptor, e.g. '(', ')'.
	// Save a matrix using cvWrite (requires: cvOpenFileStorage+cvWrite+cvReleaseFileStorage)
	//CvFileStorage* fs_m_xy = cvOpenFileStorage( "m_xy.xml", 0, CV_STORAGE_WRITE );
	//cvWrite( fs_m_xy, "m_xy", m_xy, cvAttrList(0,0) ); // cannot use non-alpha characters in object descriptor, e.g. '(', ')'.
	//cvReleaseFileStorage(&fs_m_xy);

	cvReleaseMat(&m_xy);
	
	// Load a matrix using cvLoad
	CvMat* mi_xy = (CvMat *) cvLoad( "m_xy.xml", NULL, "m_xy", NULL );	
	// Load a matrix using cvRead (requires: cvOpenFileStorage+cvGetFileNodeByName+cvRead+cvReleaseFileStorage)
	//CvFileStorage* fsi_m_xy = cvOpenFileStorage( "m_xy.xml", 0, CV_STORAGE_READ );
	//CvFileNode* node = cvGetFileNodeByName( fsi_m_xy, 0, "m_xy" );
	//CvMat* mi_xy = (CvMat *) cvRead( fsi_m_xy, node, 0);
	//cvReleaseFileStorage(&fsi_m_xy);

	// Save a matrix using cvSave
	cvSave( "mo_xy.xml", mi_xy, "m_xy", NULL, cvAttrList(0,0)); // cannot use non-alpha characters in object descriptor, e.g. '(', ')'.
    // Save a matrix using cvWrite (requires: cvOpenFileStorage+cvWrite+cvReleaseFileStorage)
	//CvFileStorage* fso_m_xy = cvOpenFileStorage( "mo_xy.xml", 0, CV_STORAGE_WRITE );
	//cvWrite( fso_m_xy, "m_xy", mi_xy, cvAttrList(0,0) ); // cannot use non-alpha characters in object descriptor, e.g. '(', ')'.
	//cvReleaseFileStorage(&fso_m_xy);

	cvReleaseMat(&mi_xy);
}


void matrixfill(CvMat *M)
{
	CvMat *P;
	P = cvCreateMat(3,2,CV_32F);
	cvRange( P, 7, P->rows+P->cols-1 );

	// CAUTION!!!
	// OpenCV automatically releases the matrix header after exiting the program function in which it was allocated.
	// This means that all data contents are lost.
	//
	// There are situations in which it is necessary to have local variables (matrices) whose dimension are
	// not known until some other function is called to determine the required size.  When these local variables
	// are defined in terms of pointers, it will be necessary to allocate some memory.  If the allocation is done
	// within some external function, the pointers are automatically released after exiting the function.
	// More specifically, the matrix header allocated with cvCreateMatHeader() is released and therefore one
	// of the following ways around the problem must be employed:
	// 1. Create an instance of CvMat rather than a CvMat* pointer.  The data stored in CvMat still needs 
	//    to be allocated as shown below, but the pointer is not realeased after exiting the function.
	// 2. Create a CvMat* pointer and allocate the header memory before calling the external function
	//    (i.e. it is assumed that the matrix header has been already allocated outside this function).
	//    If the size of the matrix is not known beforehand, simply allocate a matrix header assuming the 
	//    the matrix is of dimension 1x1 (or any non-zero dimension).
	// Notice that both approaches initialize the matrix header with the right dimensions.  Once the
	// matrix header is initialized, the space for data is (and must be) allocated.  Also notice that 
	// if the second approach is employed, then the pointer must be released in the standard way.

	// None of the following commented out lines are employed because the memory allocation of the OpenCV
	// matrix headers is lost when leaving this function.
	//	
	//M = cvCreateMat(P->rows,P->cols,CV_32F);  // Allocates a new matrix.
	//											// Equivalent to:
	//											// M = cvCreateMatHeader( P->rows, P->cols, CV_32F ); // allocate the matrix header
	//											// cvCreateData( M ); // allocate matrix data	
	//
	//M = cvCreateMatHeader( P->rows, P->cols, CV_32F ); // allocate the matrix header
														
	cvInitMatHeader( M, P->rows, P->cols, CV_32F, NULL, CV_AUTOSTEP ); // initialize the matrix header
	cvCreateData( M ); // allocate matrix data
	cvCopy(P,M,NULL);
	//M=cvCloneMat(P);

}

void matrixshow(CvMat *M)
{
	for(int row=0;row<M->rows;row++)
	{
		for(int col=0;col<M->cols;col++)
		{
            printf("M(%d,%d)=%f ",row,col,cvGetReal2D( M, row, col ));
		}
		printf("\n");
	}
}

void matrixadvanced(void)
{
	printf("Matrix 1:\n");

	CvMat M;
	matrixfill(&M);
	matrixshow(&M);

	printf("\n");


	printf("Matrix 2:\n");

	CvMat *A=cvCreateMatHeader( 1, 1, CV_32F );
	matrixfill(A);
	matrixshow(A);
	cvReleaseMat(&A);

	printf("\n");
}

int main( int argc, char** argv )
{
	// Load image in constructor: the image can be loaded either from bitmap (see cvLoadImage),
    // or from XML/YAML (see cvLoad)
    CvImage img(argc > 1 ? argv[1] : "lena.jpg", 0, CV_LOAD_IMAGE_COLOR);

    if( !img.data() ) // Check if the image has been loaded properly
	{
		fprintf( stderr, "ERROR: NULL image loaded! \n" );
		getchar();
		return -1;
	}

	// Basic Matrix Operations: Create/Save/Load
	matrixcreatesaveload();

	// Advanced Matrix Operations:  Allocating/initializing a matrix in an external function.
	// The difference between matrixcreatesaveload() and matrixadvanced() is that in the former
	// matrices are defined and allocated locally in the same matrixcreatesaveload(), while in
	// matrixadvanced() matrices are defined in the function, but initialized in the external function
	// matrixfill(), which also allocates the data space.  What should be noticed can
	// be summarized in:
	// CvMat M; // defines and creates an instance of a matrix and its associated header,
	//          // i.e the header is allocated "permanently".
	// CvMat M*=cvCreateMatHeader(...); // defines a CvMat pointer
	// In the second case the header is (and must be) allocated before calling the external function matrixfill() 
	// assuming a 1x1 (or non-zero) matrix size, even if the header will be initialized after calling the
	// external function matrixfill().
	// Read the comments within matrixadvanced() and matrixfill() for further information.
	matrixadvanced();

	// Using structures and OpenCV's CvScalar data type 
	mproc();

	// Cloning an image (create a new image with same characteristics of another image)
	img2 = img.clone(); // clone the image
    gray.create( img.size(), IPL_DEPTH_8U, 1 ); // another method to create an image - from scratch
	edge.create( img.size(), IPL_DEPTH_8U, 1 ); // another method to create an image - from scratch

	// Create a window and display the image
	cvNamedWindow( "Original Image", CV_WINDOW_AUTOSIZE );
    img.show( "Original Image" ); // .show method is the conveninient form of cvShowImage
    cvWaitKey();

	cvCvtColor( img, gray, CV_BGR2GRAY); // convert image to grayscale and store it in "gray"

	// Create another window to place the trackbar 
    cvNamedWindow(wndname, 1);
    cvNamedWindow( "Image Copy", CV_WINDOW_AUTOSIZE );

	// Create a toolbar 
    cvCreateTrackbar(tbarname, wndname, &edge_thresh, 100, on_trackbar);

    // Show the image
    on_trackbar(0);

	// Save the modified image
	cvSaveImage("test3o.jpg",img2);

    cvWaitKey();

    return 0;
    // all the images will be released automatically
}




